//------------------------------------------------------------------------
//Board Library : Seeed nRF52 Borads 1.1.8
//Board Select  : Seeed XIAO nRF52840 Sense
// 2024/06/23 @msfujino
//------------------------------------------------------------------------

#include <SPI.h>
#include <SD.h>

#include <PDM.h>
#include <RingBuf.h>    // RingBuffer by Jean-Lue - Locoduino 1.0.5

#define FIFO_SIZE   32767
#define BUFF_SIZE   32767
#define SD_CS       D2          //SD CS
#define logFileName "LOG.dat"

File logFile;
RingBuf<short, FIFO_SIZE> fifo;

short sampleBuffer[BUFF_SIZE];  // 16 bits
volatile int samplesRead;       // number of samples read

// *************************************************************** setup()
void setup() {
  NRF_POWER->DCDCEN = 1;        // enable on chip DCDC converter

  pinMode(LED_RED, OUTPUT);     // LED LOW:ON, HIGH:OFF
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_BLUE, OUTPUT);
  pinMode(D0, OUTPUT);          // for digital scope
  pinMode(D1, OUTPUT);

  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);
  digitalWrite(D0, LOW);
  digitalWrite(D1, LOW);

  if(!SD.begin(32000000UL, SD_CS)) {            // SPI clock 32MHz
    errorBlink(2);
  } 
  Serial.println(logFileName);

  logFile = SD.open(logFileName, FILE_WRITE);
  if(!logFile) { 
    errorBlink(3);
  }   

  PDM.onReceive(onPDMdata);       // configure the data receive callback
  PDM.setBufferSize(BUFF_SIZE);   // void setBufferSize(int bufferSize);
  PDM.setGain(0);                 // 0:-20dB, 40:0dB, 80:+20dB


  if (!PDM.begin(1, 16000)) {     // mono 16000 samples/sec
    errorBlink(4);
  }
}

// ************************************************************** loop()
void loop() 
{
  if (samplesRead) {                      // wait for samples to be read
    digitalWrite(D0, HIGH);
    digitalWrite(LED_GREEN, LOW);


    short temp;
    while(!fifo.isEmpty()) {
      fifo.lockedPop(temp);               // fake pop from fifo for 8kHz
      fifo.lockedPop(temp);               // pop from fifo for 8kHz
      logFile.write(byte(temp >> 8));     // save sample to the SD card
      logFile.write(byte(temp & 0x00FF));
    }
    logFile.flush();             // need "flush" to prevent file corruption
    samplesRead = 0;

    digitalWrite(LED_GREEN, HIGH);
    digitalWrite(D0, LOW);       // 400mS every 1000mS, sometimes 1500mS(need FIFO)
  }
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ onPDMdata()
void onPDMdata() 
{
  digitalWrite(D1, HIGH);
  digitalWrite(LED_BLUE, LOW);

  int bytesAvailable = PDM.available();     // the number of bytes available
  PDM.read(sampleBuffer, bytesAvailable);   // read into the sample buffer
  samplesRead = bytesAvailable / 2;         // 16-bit, 2 bytes per sample

  for(int i = 0; i < samplesRead; i++) {    // push to fifo
    if(!fifo.lockedPush(sampleBuffer[i])) {
      Serial.println("fifo FILL");
      errorBlink(5);
    }
  }

  digitalWrite(LED_BLUE, HIGH);
  digitalWrite(D1, LOW);                    // 16mS every 1000mS
}

//********************************************************************* errorBlink()
// Blinks specified number of times on error
void errorBlink(uint8_t err)
{
  Serial.println(err);
  while(true) {      
    digitalWrite(LED_GREEN, HIGH);
    digitalWrite(LED_BLUE, HIGH);
    for(int i = 0; i < err; i++) {
      digitalWrite(LED_RED, LOW);
      delay(5);
      digitalWrite(LED_RED, HIGH);
      delay(200);
    }
  delay(500);
  }
}